from typing import Any, Callable, Dict, List, Optional, Tuple

from ..logging import log
from ..model.errors import SeaplaneError
from .tasks import OpenAI, Replicate, Sql, Store, SubstationTask


class TaskEvent:
    def __init__(self, id: str, input: Any) -> None:
        self.id = id
        self.status = "in_progress"
        self.input = input
        self.output: Optional[Any] = None
        self.error: Optional[Any] = None

    def set_output(self, output: Any) -> None:
        self.output = output
        self.status = "completed"

    def set_error(self, error: Any) -> None:
        self.error = error
        self.status = "error"


SEAPLANE_API_KEY_NAME = "SEAPLANE_API_KEY"
OPENAI_API_KEY_NAME = "OPENAI_API_KEY"
REPLICATE_API_KEY_NAME = "REPLICATE_API_KEY"


class Task:
    def __init__(
        self,
        func: Callable[[Any], Any],
        type: str,
        id: Optional[str] = None,
        model: Optional[str] = None,
        sql_access: Optional[Dict[str, str]] = None,
        index_name: Optional[str] = None,
        replicas: Optional[int] = 1,
    ) -> None:
        self.func = func
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = type
        self.model = model
        self.sources: List[str] = []
        self.sql_access = sql_access
        self.sql: Optional[Sql] = None
        self.name = func.__name__
        self.index_name = index_name
        self.replicas = replicas

        if id is not None:
            self.id = id
        else:
            self.id = func.__name__

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        log.info(f"Task type '{self.type}' Model ID {self.model}")

        if self.type == "sql":
            if self.sql_access is None:
                raise SeaplaneError("Task of type SQL without sql attribute.")

            if not self.sql:
                self.sql = Sql(self.func, self.id, self.sql_access)

            return self.sql.process(*self.args, **self.kwargs)

        if self.type == "vectordb":
            log.info("Accessing Vector DB task...")

            if not self.index_name:
                raise SeaplaneError("Missing index_name attribute on vectordb Task.")

            self.args = self.args + (Store(self.index_name),)
            return self.func(*self.args, **self.kwargs)

        model = None

        if self.model:
            model = self.model.lower()
        else:
            model = self.model

        if model == "mpt-30b":
            substation = SubstationTask(self.func, self.id, model)
            return substation.process(*self.args, **self.kwargs)
        elif model == "gpt-3.5":
            openai = OpenAI(self.func, self.id, model)
            return openai.process(*self.args, **self.kwargs)
        elif model == "gpt-3":
            openai = OpenAI(self.func, self.id, model)
            return openai.process(*self.args, **self.kwargs)
        elif model == "stable-diffusion":
            replicate = Replicate(self.func, self.type, self.id, model)
            return replicate.process(*self.args, **self.kwargs)
        elif model:
            replicate = Replicate(self.func, self.type, self.id, model)
            return replicate.process(*self.args, **self.kwargs)
        else:
            log.info("Compute task type...")
            return self.func(*self.args, **self.kwargs)

    def called_from(self, sources: List[str]) -> None:
        self.sources = sources

    def print(self) -> None:
        log.info(f"id: {self.id}, type: {self.type}, model: {self.model}")



from typing import Any

from seaplane_framework.flow import processor


class TaskContext:
    """TaskContext is what a Task recieves."""

    def __init__(self, body: bytes, meta: dict[str, Any]):
        self.body = body
        # For now let's not imply that customers should touch this
        self._meta = meta

    def emit(self, message: bytes, batch_id: int = 1) -> None:
        """Queues message to send once the task completes.
        batch_id: The index with which to refer to this member of the batch."""
        new_meta = self._meta.copy()
        new_meta["_seaplane_batch_hierarchy"] += f".{batch_id}"
        output_msg = processor._Msg(message, new_meta)
        processor.write(output_msg)



from typing import Any, Callable, TypeVar

import requests
from requests import Response
import seaplane_framework.api
from seaplane_framework.api import ApiClient

from ..configuration import Configuration, config
from ..logging import log
from ..model.errors import HTTPError
from .api_http import SDK_HTTP_ERROR_CODE
from .token_api import TokenAPI

T = TypeVar("T")


def provision_req(token_api: TokenAPI) -> Callable[[Callable[[str], Response]], Any]:
    """
    Returns an object that can make token-authenticated API calls passed in as
    function parameters, mapping the results type based on `Content-Type`.

    :param token_api: the TokenAPI instance to use
    :returns: -- A function of type `((token) -> Response) -> Any` that can wrap
        token-requiring HTTP calls in a way that handles reauthentication
        correctly and parses the result based on `Content-Type`.
    """
    return lambda request: provision_token(token_api)(lambda token: _map_response(request(token)))


def provision_token(token_api: TokenAPI) -> Callable[[Callable[[str], T]], T]:
    """
    Returns an object that can make token-authenticated API calls passed in as
    function parameters.

    :param token_api: the TokenAPI instance to use
    :returns: -- A function of type `((token) -> T) -> T` that can wrap
        token-requiring functions in a way that handles reauthentication
        correctly.
    """

    def _req(request: Callable[[str], T]) -> T:
        try:
            return request(_extract_token(token_api))
        except HTTPError as e:
            _renew_if_failed(token_api, e)
            return request(_extract_token(token_api))
        except requests.exceptions.RequestException as err:
            _renew_if_failed(token_api, _map_request_exception(err))
            return request(_extract_token(token_api))

    return _req


def _map_response(response: Response) -> Any:
    """
    Extracts the response payload of an HTTP call, parsing it if possible based
    on the `Content-Type` header.

    :returns: A `bytes` payload if `Content-Type` is
        `application/octect-stream`, a JSON payload for `application/json`, and
        a `str` for anything else.
    """
    if not response.ok:
        body_error = response.text
        log.error(f"Request Error: {body_error}")
        raise HTTPError(response.status_code, body_error)
    if response.headers.get("content-type") == "application/json":
        return response.json()
    elif response.headers.get("content-type") == "application/octet-stream":
        return response.content
    else:
        return response.text


def _renew_if_failed(token_api: TokenAPI, http_error: HTTPError) -> None:
    """
    Attempts to renew a token if the given error calls for it.
    """
    if http_error.status != 401 or not token_api.auto_renew:
        raise http_error
    log.info("Auto-Renew, renewing the token...")
    token_api.renew_token()


def _extract_token(token_api: TokenAPI) -> str:
    """
    Pulls a token out of the token API, requesting one if necessary.
    """
    if token_api.access_token is None:
        token_api._request_access_token()
    return token_api.access_token or ""  # Will always be set by `_request_access_token()`


def _map_request_exception(err: requests.exceptions.RequestException) -> HTTPError:
    """
    Maps a raw `requests` exception into a Seaplane `HTTPError`.
    """
    log.error(f"Request exception: {str(err)}")
    status_code: int = SDK_HTTP_ERROR_CODE
    if err.response:
        status_code = err.response.status_code
    return HTTPError(status_code, str(err))


def get_pdk_client(access_token: str, cfg: Configuration = config) -> ApiClient:
    """
    Constructs a Seaplane PDK ApiClient from the given access token.
    """
    from .. import __version__

    pdk_config = cfg.get_platform_configuration()
    pdk_config.access_token = access_token
    client = ApiClient(pdk_config)
    client.set_default_header("X-Seaplane-Sdk-Version", __version__)
    client.set_default_header("X-Seaplane-Pdk-Version", seaplane_framework.api.__version__)
    return client



# noqa

import hashlib


def remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def file_md5(path: str) -> str:
    """
    Gets the MD5 hash of a file by path.
    """

    hasher = hashlib.md5()
    block_size = 4194304  # 4 MB
    with open(path, "rb") as fh:
        while True:
            buffer = fh.read(block_size)
            if not buffer:
                break
            hasher.update(buffer)
    return hasher.hexdigest()



from typing import Any, Callable, Dict, Optional, Tuple

import requests

from ...api.api_http import SDK_HTTP_ERROR_CODE
from ...configuration import config
from ...logging import log
from ...model.errors import HTTPError, SeaplaneError

OPENAI_API_KEY_NAME = "OPENAI_API_KEY"
OPENAI_URL = "https://api.openai.com"


def _openai_headers() -> Dict[str, str]:
    return {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "Authorization": f"Bearer {config._api_keys[OPENAI_API_KEY_NAME]}",
    }


def _post(url: str, json: Any) -> Any:
    try:
        response = requests.post(
            url,
            json=json,
            headers=_openai_headers(),
        )
        if response.ok:
            return response.json()
        else:
            body_error = response.text
            log.error(f"Request Error: {body_error}")
            raise HTTPError(response.status_code, body_error)

    except requests.exceptions.RequestException as err:
        log.error(f"Request exception: {str(err)}")
        raise HTTPError(SDK_HTTP_ERROR_CODE, str(err))


def _check_openai_api_keys() -> None:
    if config._api_keys is None:
        raise SeaplaneError(
            f"OPENAI API Key `{OPENAI_API_KEY_NAME}` is not set, \
                  use `sea.config.set_api_keys`."
        )
    elif not config._api_keys.get(OPENAI_API_KEY_NAME, None):
        raise SeaplaneError(
            f"OPENAI API Key `{OPENAI_API_KEY_NAME}` is not set,\
                  use `sea.config.set_api_keys`."
        )


def _gpt_3() -> Callable[[Dict[str, Any]], Any]:
    def model(payload: Dict[str, Any]) -> Any:
        """
        The OpenAI GPT-3 payload.

        payload = {"model": "text-davinci-003", "prompt": prompt, "temperature": 0}
        """
        _check_openai_api_keys()

        url = f"{OPENAI_URL}/v1/completions"

        return _post(
            url,
            json=payload,
        )

    return model


def _gpt_3_5() -> Callable[[Dict[str, Any]], Any]:
    def model(payload: Dict[str, Any]) -> Any:
        """
        The OpenAI GPT-3.5 payload.

         payload = {
           "model": "gpt-3.5-turbo",
           "messages": [{"role": "user", "content": "Any prompt here.."}],
           "temperature": 0.7,
         }
        """
        _check_openai_api_keys()

        url = f"{OPENAI_URL}/v1/chat/completions"

        return _post(
            url,
            json=payload,
        )

    return model


class OpenAI:
    def __init__(self, func: Callable[[Any], Any], id: str, model: Optional[str]) -> None:
        self.func = func
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = "inference"
        self.model = model
        self.id = id

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        if self.model == "gpt-3.5":
            log.info("Processing GPT-3.5 Model...")
            self.args = self.args + (_gpt_3_5(),)

            return self.func(*self.args, **self.kwargs)
        elif self.model == "gpt-3":
            log.info("Processing GPT-3 Model...")
            self.args = self.args + (_gpt_3(),)

            return self.func(*self.args, **self.kwargs)
        else:
            raise SeaplaneError(
                "OPENAI Task Not recognised , the model \
                    doesn't match with any current OPEN AI task."
            )

    def print(self) -> None:
        log.info(f"id: {self.id}, type: {self.type}, model: {self.model}")



import traceback
from typing import Any, Callable, Dict, List, Optional

from .app import AppEvent
from .task import TaskEvent


def format_exception(e: Optional[Exception]) -> Optional[List[str]]:
    if e is None:
        return None

    return traceback.format_exception(type(e), e, e.__traceback__)


def task_event_json(event: TaskEvent) -> Dict[str, Any]:
    return {
        "id": event.id,
        "input": event.input,
        "status": event.status,
        "output": event.output,
        "error": format_exception(event.error),
    }


def event_json(event: AppEvent) -> Dict[str, Any]:
    return {
        "id": event.id,
        "app_id": event.app_id,
        "input": event.input,
        "status": event.status,
        "output": event.output,
        "error": format_exception(event.error),
        "tasks": [task_event_json(task) for task in event.tasks],
    }


class EventHandler:
    def __init__(self) -> None:
        self.on_change_event: Callable[[Dict[str, Any]], None] = lambda x: None
        self.events: List[AppEvent] = []
        self.active_event: List[str] = ["none"]

    def set_event(self, callback: Callable[[Dict[str, Any]], None]) -> None:
        self.on_change_event = callback

    def on_change(self, message: Dict[str, Any]) -> None:
        self.on_change_event(message)

    def add_event(self, event: AppEvent) -> None:
        self.active_event[0] = event.id
        self.events.append(event)

        self.on_change_event({"type": "add_request", "payload": event_json(event)})

    def update_event(self, event: AppEvent) -> None:
        for i, e in enumerate(self.events):
            if e.id == self.active_event[0]:
                self.events[i] = event

                self.on_change_event({"type": "update_request", "payload": event_json(event)})
                break

    def task_event(self, task_event: TaskEvent) -> None:
        for i, event in enumerate(self.events):
            if event.id == self.active_event[0]:
                event.add_task_event(task_event)

                self.events[i] = event

                self.on_change_event({"type": "update_request", "payload": event_json(event)})
                break



import json
from typing import Any, Optional

import requests

from ..configuration import Configuration, config
from .api_http import headers
from .api_request import provision_req


class SubstationAPI:
    """
    Class for handle Substation API calls.
    """

    def __init__(self, configuration: Configuration = config) -> None:
        self.url = f"{configuration.substation_endpoint}/completions"
        self.req = provision_req(configuration._token_api)

    # This is the template that works best with the mpt-30b-instruct model;
    # strongly advised to use this format.
    # TODO: Generalize this to support other models
    def _format_prompt(self, instruction: str) -> str:
        template = "Below is an instruction that describes a task. Write a response that appropriately completes the request.\n\n###Instruction\n{instruction}\n\n### Response\n"  # noqa
        return template.format(instruction=instruction)

    def predict(
        self,
        prompt: str,
        max_output_length: Optional[int] = 3000,
        model_specific_prompt: Optional[bool] = True,
    ) -> Any:
        if model_specific_prompt:
            prompt = self._format_prompt(prompt)
        result = self.req(
            lambda access_token: requests.post(
                self.url,
                headers=headers(access_token),
                json={"max_tokens": max_output_length, "prompt": prompt},
            )
        )

        if (
            type(result) is str
        ):  # Substation FIX, It needs to returns content-type: application/json
            return json.loads(result)
        else:
            return result



"""
command line interface
"""
from argparse import ArgumentParser
from importlib.metadata import version
import os
from typing import Any, Dict

from cookiecutter.main import cookiecutter

from .build import read_project_file, validate_project


def get_project() -> Dict[str, Any]:
    validate_project()
    return read_project_file()


def cli_deploy(task: str) -> None:
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} deploy {task}")


def cli_destroy() -> None:
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} destroy")


def cli_build() -> None:
    project = get_project()
    project_name = project["tool"]["poetry"]["name"]
    main_script = project["tool"]["seaplane"]["main"]
    os.system(f"poetry run python {project_name}/{main_script} build")


def init(project_name: str) -> None:
    cookiecutter_template = "https://github.com/seaplane-io/seaplane-app-python-template.git"
    project_directory = "."

    extra_context = {"project_slug": project_name, "seaplane_version": version("seaplane")}

    cookiecutter(
        cookiecutter_template,
        output_dir=project_directory,
        no_input=True,  # Disable any interactive prompts
        extra_context=extra_context,
    )

    print(f"🛩️  {project_name} project generated successfully!")


def main() -> None:
    parser = ArgumentParser(prog="seaplane", description="Seaplane Apps command line interface")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Build command
    subparsers.add_parser("build", help="Build command")

    # Deploy command
    deploy_parser = subparsers.add_parser("deploy", help="Deploy command")
    deploy_parser.add_argument(
        "--task",
        type=str,
        help="seaplane deploy which can include a TASK_ID as a parameter to deploy an individual TASK",  # noqa
    )

    # Destroy command
    subparsers.add_parser("destroy", help="Remove your App and associated tasks")

    # Init command
    init_parser = subparsers.add_parser("init", help="Init command")
    init_parser.add_argument("app", help="Seaplane Apps name")

    # Version argument
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=version("seaplane")),
    )

    args = parser.parse_args()

    if args.command == "build":
        cli_build()

    elif args.command == "deploy":
        task = ""
        if args.task:
            task = args.task

        cli_deploy(task)

    elif args.command == "destroy":
        cli_destroy()

    elif args.command == "init":
        init(args.app)

    else:
        parser.print_help()

    exit(0)
